import operator as op

import numpy as onp
from numpy_utils import binary_ufunc_check, binary_ufunc_check_no_same_args, stat_check, unary_ufunc_check

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd.test_util import combo_check

npr.seed(0)


# Array statistics functions
def test_max():
    stat_check(np.max)


# def test_all():  stat_check(np.all)
# def test_any():  stat_check(np.any)
def test_max():
    stat_check(np.max)


def test_mean():
    stat_check(np.mean)


def test_min():
    stat_check(np.min)


def test_sum():
    stat_check(np.sum)


def test_prod():
    stat_check(np.prod)


def test_var():
    stat_check(np.var)


def test_std():
    stat_check(np.std)


# Unary ufunc tests


def test_sin():
    unary_ufunc_check(np.sin)


def test_abs():
    unary_ufunc_check(np.abs, lims=[0.1, 4.0])


def test_absolute():
    unary_ufunc_check(np.absolute, lims=[0.1, 4.0])


def test_arccosh():
    unary_ufunc_check(np.arccosh, lims=[1.1, 4.0])


def test_arcsinh():
    unary_ufunc_check(np.arcsinh, lims=[-0.9, 0.9])


def test_arctanh():
    unary_ufunc_check(np.arctanh, lims=[-0.9, 0.9])


def test_ceil():
    unary_ufunc_check(np.ceil, lims=[-1.5, 1.5], test_complex=False)


def test_cos():
    unary_ufunc_check(np.cos)


def test_cosh():
    unary_ufunc_check(np.cosh)


def test_deg2rad():
    unary_ufunc_check(np.deg2rad, test_complex=False)


def test_degrees():
    unary_ufunc_check(lambda x: np.degrees(x) / 50.0, test_complex=False)


def test_exp():
    unary_ufunc_check(np.exp)


def test_exp2():
    unary_ufunc_check(np.exp2)


def test_expm1():
    unary_ufunc_check(np.expm1)


def test_fabs():
    unary_ufunc_check(np.fabs, test_complex=False)


def test_floor():
    unary_ufunc_check(np.floor, lims=[-1.5, 1.5], test_complex=False)


def test_log():
    unary_ufunc_check(np.log, lims=[0.2, 2.0])


def test_log10():
    unary_ufunc_check(np.log10, lims=[0.2, 2.0])


def test_log1p():
    unary_ufunc_check(np.log1p, lims=[0.2, 2.0])


def test_log2():
    unary_ufunc_check(np.log2, lims=[0.2, 2.0])


def test_rad2deg():
    unary_ufunc_check(lambda x: np.rad2deg(x) / 50.0, test_complex=False)


def test_radians():
    unary_ufunc_check(np.radians, test_complex=False)


def test_sign():
    unary_ufunc_check(np.sign, test_complex=False)


def test_sin():
    unary_ufunc_check(np.sin)


def test_sinh():
    unary_ufunc_check(np.sinh)


def test_sqrt():
    unary_ufunc_check(np.sqrt, lims=[1.0, 3.0])


def test_square():
    unary_ufunc_check(np.square, test_complex=False)


def test_tan():
    unary_ufunc_check(np.tan, lims=[-1.1, 1.1])


def test_tanh():
    unary_ufunc_check(np.tanh)


def test_real():
    unary_ufunc_check(np.real)


def test_real_ic():
    unary_ufunc_check(np.real_if_close)


def test_imag():
    unary_ufunc_check(np.imag)


def test_conj():
    unary_ufunc_check(np.conj)


def test_angle():
    unary_ufunc_check(np.angle)


# Binary ufunc tests


def test_add():
    binary_ufunc_check(np.add)


def test_logaddexp():
    binary_ufunc_check(np.logaddexp, test_complex=False)


def test_logaddexp2():
    binary_ufunc_check(np.logaddexp2, test_complex=False)


def test_remainder():
    binary_ufunc_check_no_same_args(np.remainder, lims_A=[-0.9, 0.9], lims_B=[0.7, 1.9], test_complex=False)


def test_true_divide():
    binary_ufunc_check(np.true_divide, lims_B=[0.8, 1.2], test_complex=False)


def test_mod():
    binary_ufunc_check_no_same_args(np.mod, lims_B=[0.8, 2.1], test_complex=False)


def test_true_divide_neg():
    binary_ufunc_check(np.true_divide, lims_B=[-0.3, -2.0], test_complex=False)


def test_mod_neg():
    binary_ufunc_check_no_same_args(np.mod, lims_B=[-0.3, -2.0], test_complex=False)


def test_op_mul():
    binary_ufunc_check(op.mul)


def test_op_add():
    binary_ufunc_check(op.add)


def test_op_sub():
    binary_ufunc_check(op.sub)


def test_op_mod():
    binary_ufunc_check_no_same_args(op.mod, lims_B=[0.3, 2.0], test_complex=False)


def test_op_mod_neg():
    binary_ufunc_check_no_same_args(op.mod, lims_B=[-0.3, -2.0], test_complex=False)


# Misc tests

R = npr.randn
C = lambda *shape: npr.randn(*shape) + 1j * npr.randn(*shape)


def test_transpose():
    combo_check(np.transpose, [0])(
        [R(2, 3, 4)], axes=[None, [0, 1, 2], [0, 2, 1], [2, 0, 1], [2, 1, 0], [1, 0, 2], [1, 2, 0]]
    )


def test_moveaxis():
    combo_check(np.moveaxis, [0])([R(2, 3, 4)], source=[0, 1, 2], destination=[0, 1, 2])


def test_repeat():
    combo_check(np.repeat, [0])([R(2, 3, 4), R(3, 1)], repeats=[0, 1, 2], axis=[None, 0, 1])


def test_diff():
    combo_check(np.diff, [0])([R(5, 5), R(5, 5, 5)], n=[1, 2], axis=[0, 1])
    combo_check(np.diff, [0])([R(1), R(1, 1)], axis=[0])
    combo_check(np.diff, [0])([R(1, 1), R(3, 1)], axis=[1])


def test_gradient():
    combo_check(np.gradient, [0])([R(5, 5), R(5, 5, 5)], axis=[None, 0, 1, -1])
    combo_check(np.gradient, [0])([R(5, 5, 5)], axis=[(0, 1), (0, -1)])


def test_tile():
    combo_check(np.tile, [0])([R(2, 1, 3, 1)], reps=[(1, 4, 1, 2)])
    combo_check(np.tile, [0])([R(1, 2)], reps=[(1, 2), (2, 3), (3, 2, 1)])
    combo_check(np.tile, [0])([R(1)], reps=[(2,), 2])


def test_kron():
    combo_check(np.kron, [0, 1])(
        [R(5, 5), R(4, 4), R(5), R(5, 1), R(1, 5), R(), C(5, 5)],
        [R(3, 3), R(2, 2), R(3), R(1, 3), R(3, 1), R(), C(3, 3)],
    )


def test_inner():
    combo_check(np.inner, [0, 1])([1.5, R(3), R(2, 3)], [0.3, R(3), R(4, 3)])


def test_dot():
    combo_check(np.dot, [0, 1], order=3)(
        [1.5, R(3), R(2, 3), R(2, 2, 3), C(3), C(2, 3)], [0.3, R(3), R(3, 4), R(2, 3, 4), C(3)]
    )


def test_outer():
    combo_check(np.outer, [0, 1], order=3)([R(3), C(3)], [R(3), C(3)])


def test_matmul():
    combo_check(np.matmul, [0, 1])(
        [R(3), R(2, 3), R(2, 2, 3), C(3), C(2, 3)], [R(3), R(3, 4), R(2, 3, 4), C(3), C(3, 4)]
    )


def test_matmul_broadcast():
    combo_check(np.matmul, [0, 1])([R(1, 2, 2)], [R(3, 2, 1)])


def test_tensordot_1():
    combo_check(np.tensordot, [0, 1], order=3)(
        [R(1, 3), R(2, 3, 2), C(1, 3)], [R(3), R(3, 1), R(3, 4, 2), C(3)], axes=[[(1,), (0,)]]
    )


def test_tensordot_2():
    combo_check(np.tensordot, [0, 1], order=3)(
        [R(3), R(3, 1), R(3, 4, 2)], [R(1, 3), R(2, 3, 2)], axes=[[(0,), (1,)]]
    )


def test_tensordot_3():
    combo_check(np.tensordot, [0, 1], order=3)(
        [R(2, 3), R(2, 3, 4)], [R(1, 2, 3), R(2, 2, 3, 4)], axes=[[(0, 1), (1, 2)], [(1, 0), (2, 1)]]
    )


def test_tensordot_4():
    combo_check(np.tensordot, [0, 1], order=3)([R(2, 2), R(4, 2, 2)], [R(2, 2), R(2, 2, 4)], axes=[1, 2])


def test_tensordot_5():
    combo_check(np.tensordot, [0, 1], order=3)([R(4)], [R()], axes=[0])


def test_tensordot_6():
    combo_check(np.tensordot, [0, 1], order=3)([R(2, 6)], [R(6, 3)], axes=[[[-1], [0]]])


def test_tensordot_7():
    combo_check(np.tensordot, [0, 1], order=3)([R(2, 6)], [R(6, 3)], axes=[[-1, 0]])


def test_tensordot_8():
    combo_check(np.tensordot, [0, 1], order=3)([R(2)], [R(2, 2)], axes=[[0, 1]])


# Need custom tests because gradient is undefined when arguments are identical.
def test_maximum():
    combo_check(np.maximum, [0, 1])([R(1), R(1, 4), R(3, 4)], [R(1), R(1, 4), R(3, 4)])


def test_fmax():
    combo_check(np.fmax, [0, 1])([R(1), R(1, 4), R(3, 4)], [R(1), R(1, 4), R(3, 4)])


def test_minimum():
    combo_check(np.minimum, [0, 1])([R(1), R(1, 4), R(3, 4)], [R(1), R(1, 4), R(3, 4)])


def test_fmin():
    combo_check(np.fmin, [0, 1])([R(1), R(1, 4), R(3, 4)], [R(1), R(1, 4), R(3, 4)])


def test_sort():
    combo_check(np.sort, [0])([R(1), R(7)])


if onp.lib.NumpyVersion(onp.__version__) < "2.0.0":

    def test_msort():
        combo_check(np.msort, [0])([R(1), R(7)])


def test_partition():
    combo_check(np.partition, [0])([R(7), R(14)], kth=[0, 3, 6])


def test_atleast_1d():
    combo_check(np.atleast_1d, [0])([1.2, R(1), R(7), R(1, 4), R(2, 4), R(2, 4, 5)])


def test_atleast_2d():
    combo_check(np.atleast_2d, [0])([1.2, R(1), R(7), R(1, 4), R(2, 4), R(2, 4, 5)])


def test_atleast_3d():
    combo_check(np.atleast_3d, [0])([1.2, R(1), R(7), R(1, 4), R(2, 4), R(2, 4, 5), R(2, 4, 3, 5)])


def test_einsum_transpose():
    combo_check(np.einsum, [1])(["ij->ji"], [R(1, 1), R(4, 4), R(3, 4)])


def test_einsum_matmult():
    combo_check(np.einsum, [1, 2])(["ij,jk->ik"], [R(2, 3), C(2, 3)], [R(3, 4), C(3, 4)])


def test_einsum_matmult_broadcast():
    combo_check(np.einsum, [1, 2])(["...ij,...jk->...ik"], [R(2, 3), R(2, 2, 3)], [R(3, 4), R(2, 3, 4)])


def test_einsum_matmult_broadcast_leadzero():
    combo_check(np.einsum, [1, 2])(["...ij,...jk->...ik"], [R(0, 2, 3)], [R(0, 3, 4)])


def test_einsum_covsum():
    combo_check(np.einsum, [1, 2])(["ijk,lji->lki"], [R(3, 4, 4)], [R(4, 4, 3)])


def test_einsum_ellipses():
    combo_check(np.einsum, [1, 2])(
        ["...jk,...lj->...lk", "...,...->..."], [R(4, 4), R(3, 4, 4)], [R(4, 4), R(3, 4, 4)]
    )


def test_einsum_ellipses_tail():
    combo_check(np.einsum, [1, 2])(["jk...,lj...->lk..."], [R(3, 2), R(3, 2, 4)], [R(2, 3), R(2, 3, 4)])


def test_einsum_ellipses_center():
    combo_check(np.einsum, [1, 2])(["j...k,lj...->lk..."], [R(2, 2), R(2, 2, 2)], [R(2, 2), R(2, 2, 2)])


def test_einsum_three_args():
    combo_check(np.einsum, [1, 2])(["ijk,lji,lli->lki"], [R(3, 4, 4)], [R(4, 4, 3)], [R(4, 4, 3)])


def test_einsum2_transpose():
    combo_check(np.einsum, [0])([R(1, 1), R(4, 4), R(3, 4)], [(0, 1)], [(1, 0)])


def test_einsum2_matmult():
    combo_check(np.einsum, [0, 2])([R(2, 3)], [(0, 1)], [R(3, 4)], [(1, 2)], [(0, 2)])


def test_einsum2_matmult_broadcast():
    combo_check(np.einsum, [0, 2])(
        [R(2, 3), R(2, 2, 3)],
        [(Ellipsis, 0, 1)],
        [R(3, 4), R(2, 3, 4)],
        [(Ellipsis, 1, 2)],
        [(Ellipsis, 0, 2)],
    )


def test_einsum2_covsum():
    combo_check(np.einsum, [0, 2])([R(3, 4, 4)], [(0, 1, 2)], [R(4, 4, 3)], [(3, 1, 0)], [(3, 2, 0)])


def test_einsum2_three_args():
    combo_check(np.einsum, [0, 2])(
        [R(3, 4, 4)], [(0, 1, 2)], [R(4, 4, 3)], [(3, 1, 0)], [R(4, 4, 3)], [(3, 3, 0)], [(3, 2, 0)]
    )


def test_einsum_naked_sum():
    combo_check(np.einsum, [1, 2])(["k,nk->"], [R(5)], [R(10, 5)])


def test_einsum_naked_sum2():
    combo_check(np.einsum, [1])(["abcd->bd"], [R(3, 2, 3, 2)])


def test_einsum_naked_sum_ellipsis():
    combo_check(np.einsum, [1, 2])(["...k,...nk->..."], [R(3, 5)], [R(3, 10, 5)])


def test_einsum_no_output_indices():
    combo_check(np.einsum, [1, 2])(["ij,k"], [R(3, 4)], [R(3)])


def test_trace():
    combo_check(np.trace, [0])([R(5, 5), R(4, 5), R(5, 4), R(3, 4, 5)], offset=[-1, 0, 1])


def test_diag():
    combo_check(np.diag, [0])([R(5, 5)], k=[-1, 0, 1])


def test_diag_flat():
    combo_check(np.diag, [0])([R(5)], k=[-1, 0, 1])


def test_tril():
    combo_check(np.tril, [0])([R(5, 5)], k=[-1, 0, 1])


def test_triu():
    combo_check(np.triu, [0])([R(5, 5)], k=[-1, 0, 1])


def test_tril_3d():
    combo_check(np.tril, [0])([R(5, 5, 4)], k=[-1, 0, 1])


def test_triu_3d():
    combo_check(np.triu, [0])([R(5, 5, 4)], k=[-1, 0, 1])


def test_swapaxes():
    combo_check(np.swapaxes, [0])([R(3, 4, 5)], axis1=[0, 1, 2], axis2=[0, 1, 2])


def test_rollaxis():
    combo_check(np.rollaxis, [0])([R(2, 3, 4)], axis=[0, 1, 2], start=[0, 1, 2])


def test_cross():
    combo_check(np.cross, [0, 1])(
        [R(3, 3)], [R(3, 3)], axisa=[-1, 0, 1], axisb=[-1, 0, 1], axisc=[-1, 0, 1], axis=[None, -1, 0, 1]
    )


def test_vsplit_2d():
    combo_check(np.vsplit, [0])([R(4, 8)], [4, [1, 2]])


def test_vsplit_3d():
    combo_check(np.vsplit, [0])([R(4, 4, 4)], [2, [1, 2]])


def test_hsplit_2d():
    combo_check(np.hsplit, [0])([R(4, 8)], [4, [1, 2]])


def test_hsplit_3d():
    combo_check(np.hsplit, [0])([R(4, 4, 4)], [2, [1, 2]])


def test_dsplit_3d():
    combo_check(np.dsplit, [0])([R(4, 4, 4)], [2, [1, 2]])


def test_split_1d():
    combo_check(np.split, [0])([R(1), R(7)], [1], axis=[0])


def test_split_2d():
    combo_check(np.split, [0])([R(4, 8)], [4, [1, 2]], axis=[0, 1])


def test_split_3d():
    combo_check(np.split, [0])([R(4, 4, 4)], [2, [1, 2]], axis=[0, 1, 2])


def test_array_split_1d():
    combo_check(np.array_split, [0])([R(1), R(7)], [1, 3], axis=[0])


def test_array_split_2d():
    combo_check(np.array_split, [0])([R(7, 7)], [4, [3, 5]], axis=[0, 1])


def test_array_split_3d():
    combo_check(np.array_split, [0])([R(7, 7, 7)], [4, [3, 5]], axis=[0, 1, 2])


def test_concatenate_1ist():
    combo_check(np.concatenate, [0])([(R(1), R(3))], axis=[0])


def test_concatenate_tuple():
    combo_check(np.concatenate, [0])([[R(1), R(3)]], axis=[0])


def test_concatenate_2d():
    combo_check(np.concatenate, [0])([(R(2, 2), R(2, 2))], axis=[0, 1])


def test_concatenate_3d():
    combo_check(np.concatenate, [0])([(R(2, 2, 2), R(2, 2, 2))], axis=[0, 1, 2])


def test_vstack_1d():
    combo_check(np.vstack, [0])([R(2), (R(2), R(2))])


def test_vstack_2d():
    combo_check(np.vstack, [0])([R(2, 3), (R(2, 4), R(1, 4))])


def test_vstack_3d():
    combo_check(np.vstack, [0])([R(2, 3, 4), (R(2, 3, 4), R(5, 3, 4))])


def test_hstack_1d():
    combo_check(np.hstack, [0])([R(2), (R(2), R(2))])


def test_hstack_2d():
    combo_check(np.hstack, [0])([R(3, 2), (R(3, 4), R(3, 5))])


def test_hstack_3d():
    combo_check(np.hstack, [0])([R(2, 3, 4), (R(2, 1, 4), R(2, 5, 4))])


def test_stack_1d():
    combo_check(np.stack, [0])([(R(2),), (R(2), R(2))], axis=[0, 1])


def test_row_stack_1d():
    combo_check(np.row_stack, [0])([R(2), (R(2), R(2))])


def test_row_stack_2d():
    combo_check(np.row_stack, [0])([R(2, 3), (R(2, 4), R(1, 4))])


def test_column_stack_1d():
    combo_check(np.column_stack, [0])([R(2), (R(2), R(2))])


def test_column_stack_2d():
    combo_check(np.column_stack, [0])([R(2, 2), (R(2, 2), R(2, 2))])


def test_select():
    combo_check(np.select, [1])(
        [[R(3, 4, 5) > 0, R(3, 4, 5) > 0, R(3, 4, 5) > 0]],
        [[R(3, 4, 5), R(3, 4, 5), R(3, 4, 5)]],
        default=[0.0, 1.1],
    )


def test_pad():
    combo_check(np.pad, [0])(
        [R(2, 2)], [0, 3, (3,), (3, 2), ((3, 2),), ((1, 2), (3, 4)), ((0, 0), (0, 0))], ["constant"]
    )



import warnings
from functools import partial

import pytest

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import (
    checkpoint,
    elementwise_grad,
    grad,
    grad_and_aux,
    hessian,
    hessian_tensor_product,
    jacobian,
    make_ggnvp,
    make_hvp,
    make_jvp,
    tensor_jacobian_product,
    value_and_grad,
)
from autograd.test_util import check_equivalent, check_grads  # , nd
from autograd.tracer import isbox

npr.seed(1)


def test_return_both():
    fun = lambda x: 3.0 * x**3.2
    d_fun = grad(fun)
    f_and_d_fun = value_and_grad(fun)

    test_x = 1.7
    f, d = f_and_d_fun(test_x)
    assert f == fun(test_x)
    assert d == d_fun(test_x)


def test_value_and_grad():
    fun = lambda x: np.sum(np.sin(x) ** 2)
    dfun = grad(fun)
    dfun_both = value_and_grad(fun)
    x = npr.randn(5)
    assert not isbox(dfun_both(x)[0])
    check_equivalent(fun(x), dfun_both(x)[0])
    check_equivalent(dfun(x), dfun_both(x)[1])

    def fun2(x):
        return dfun_both(x)[0]

    check_grads(fun2)(x)


def test_hessian():
    # Check Hessian of a quadratic function.
    D = 5
    H = npr.randn(D, D)

    def fun(x):
        return np.dot(np.dot(x, H), x)

    hess = hessian(fun)
    x = npr.randn(D)
    check_equivalent(hess(x), H + H.T)


def test_multigrad():
    def complicated_fun(a, b, c, d, e, f=1.1, g=9.0):
        return a + np.sin(b) + np.cosh(c) + np.cos(d) + np.tan(e) + f + g

    def complicated_fun_3_1(d_b):
        d, b = d_b
        return complicated_fun(A, b, C, d, E, f=F, g=G)

    A = 0.5
    B = -0.3
    C = 0.2
    D = -1.1
    E = 0.7
    F = 0.6
    G = -0.1

    wrapped = grad(complicated_fun, argnum=[3, 1])(A, B, C, D, E, f=F, g=G)
    explicit = grad(complicated_fun_3_1)((D, B))
    check_equivalent(wrapped, explicit)


def test_value_and_multigrad():
    def complicated_fun(a, b, c, d, e, f=1.1, g=9.0):
        return a + np.sin(b) + np.cosh(c) + np.cos(d) + np.tan(e) + f + g

    A = 0.5
    B = -0.3
    C = 0.2
    D = -1.1
    E = 0.7
    F = 0.6
    G = -0.1

    dfun = grad(complicated_fun, argnum=[3, 1])
    dfun_both = value_and_grad(complicated_fun, argnum=[3, 1])

    check_equivalent(complicated_fun(A, B, C, D, E, f=F, g=G), dfun_both(A, B, C, D, E, f=F, g=G)[0])

    check_equivalent(dfun(A, B, C, D, E, f=F, g=G), dfun_both(A, B, C, D, E, f=F, g=G)[1])


def test_multigrad_onearg():
    fun = lambda x, y: np.sum(x + np.sin(y))
    packed_fun = lambda xy: np.sum(xy[0] + np.sin(xy[1]))
    A, B = npr.randn(3), npr.randn(3)
    check_equivalent(grad(fun, argnum=[0])(A, B), (grad(packed_fun)((A, B))[0],))


def test_elementwise_grad():
    def simple_fun(a):
        return a + np.sin(a) + np.cosh(a)

    A = npr.randn(10)

    wrapped = elementwise_grad(simple_fun)(A)
    explicit = np.array([grad(simple_fun)(A[i]) for i in range(len(A))])
    check_equivalent(wrapped, explicit)


def test_elementwise_grad_multiple_args():
    def simple_fun(a, b):
        return a + np.sin(a) + np.cosh(b)

    A = 0.9
    B = npr.randn(10)
    argnum = 1

    wrapped = elementwise_grad(simple_fun, argnum)(A, B)
    explicit = np.array([grad(simple_fun, argnum)(A, B[i]) for i in range(len(B))])
    check_equivalent(wrapped, explicit)


def test_hessian_tensor_product():
    fun = lambda a: np.sum(np.sin(a))
    a = npr.randn(5)
    v = npr.randn(5)
    H = hessian(fun)(a)
    check_equivalent(np.dot(H, v), hessian_tensor_product(fun)(a, v))


def test_hvp():
    fun = lambda a: np.sum(np.sin(a))
    a = npr.randn(5)
    v = npr.randn(5)
    H = hessian(fun)(a)
    hvp = make_hvp(fun)(a)[0]
    check_equivalent(np.dot(H, v), hvp(v))


def test_hessian_matrix_product():
    fun = lambda a: np.sum(np.sin(a))
    a = npr.randn(5, 4)
    V = npr.randn(5, 4)
    H = hessian(fun)(a)
    check_equivalent(np.tensordot(H, V), hessian_tensor_product(fun)(a, V))


def test_hessian_tensor_product_3d():
    fun = lambda a: np.sum(np.sin(a))
    a = npr.randn(5, 4, 3)
    V = npr.randn(5, 4, 3)
    H = hessian(fun)(a)
    check_equivalent(np.tensordot(H, V, axes=np.ndim(V)), hessian_tensor_product(fun)(a, V))


def test_tensor_jacobian_product():
    # This function will have an asymmetric jacobian matrix.
    fun = lambda a: np.roll(np.sin(a), 1)
    a = npr.randn(5)
    V = npr.randn(5)
    J = jacobian(fun)(a)
    check_equivalent(np.dot(V.T, J), tensor_jacobian_product(fun)(a, V))


def test_matrix_jacobian_product():
    fun = lambda a: np.roll(np.sin(a), 1)
    a = npr.randn(5, 4)
    V = npr.randn(5, 4)
    J = jacobian(fun)(a)
    check_equivalent(np.tensordot(V, J), tensor_jacobian_product(fun)(a, V))


def test_tensor_jacobian_product():
    fun = lambda a: np.roll(np.sin(a), 1)
    a = npr.randn(5, 4, 3)
    V = npr.randn(5, 4)
    J = jacobian(fun)(a)
    check_equivalent(np.tensordot(V, J, axes=np.ndim(V)), tensor_jacobian_product(fun)(a, V))


def test_deprecated_defgrad_wrapper():
    from autograd.core import primitive

    @primitive
    def new_mul(x, y):
        return x * y

    with warnings.catch_warnings(record=True) as w:
        new_mul.defgrad(lambda ans, x, y: lambda g: y * g)
        new_mul.defgrad(lambda ans, x, y: lambda g: x * g, argnum=1)

    def fun(x, y):
        return new_mul(x, y)

    mat1 = npr.randn(2, 2)
    mat2 = npr.randn(2, 2)
    check_grads(fun, modes=["rev"])(mat1, mat2)


def test_deprecated_defvjp_wrapper():
    from autograd.core import primitive

    @primitive
    def new_mul(x, y):
        return x * y

    with warnings.catch_warnings(record=True) as w:
        new_mul.defvjp(lambda g, ans, vs, gvs, x, y: y * g)
        new_mul.defvjp(lambda g, ans, vs, gvs, x, y: x * g, argnum=1)

    def fun(x, y):
        return new_mul(x, y)

    mat1 = npr.randn(2, 2)
    mat2 = npr.randn(2, 2)
    check_grads(fun, modes=["rev"])(mat1, mat2)


def test_deprecated_defvjp_is_zero_wrapper():
    from autograd.core import primitive

    @primitive
    def new_mul(x, y):
        return 0 * x * y

    with warnings.catch_warnings(record=True) as w:
        new_mul.defvjp_is_zero([0, 1])

    def fun(x, y):
        return new_mul(x, y)

    mat1 = npr.randn(2, 2)
    mat2 = npr.randn(2, 2)
    with warnings.catch_warnings(record=True) as w:
        check_grads(fun, modes=["rev"])(mat1, mat2)


def test_deprecated_quick_grad_check_wrapper():
    from autograd.util import quick_grad_check

    with warnings.catch_warnings(record=True) as w:
        quick_grad_check(lambda x, y: x**2 + y, 1.0, (2.0,))


def test_partial():
    def f(x, y):
        return x

    grad(partial(f, y=1))


@pytest.mark.skip(reason="fails with NumPy nightlies")
def test_dtypes():
    def f(x):
        return np.real(np.sum(x**2))

    # Array y with dtype np.float32
    y = np.random.randn(10, 10).astype(np.float32)
    assert grad(f)(y).dtype.type is np.float32

    y = np.random.randn(10, 10).astype(np.float16)
    assert grad(f)(y).dtype.type is np.float16

    y = np.random.randn(10, 10).astype(np.longdouble)
    grad(f)(y)

    y = np.random.randn(10, 10).astype(np.clongdouble)
    grad(f)(y)


def test_checkpoint_correctness():
    bar = lambda x, y: 2 * x + y + 5
    checkpointed_bar = checkpoint(bar)
    foo = lambda x: bar(x, x / 3.0) + bar(x, x**2)
    foo2 = lambda x: checkpointed_bar(x, x / 3.0) + checkpointed_bar(x, x**2)
    assert np.allclose(foo(3.0), foo2(3.0))
    assert np.allclose(grad(foo)(3.0), grad(foo2)(3.0))

    baz = lambda *args: sum(args)
    checkpointed_baz = checkpoint(baz)
    foobaz = lambda x: baz(x, x / 3.0)
    foobaz2 = lambda x: checkpointed_baz(x, x / 3.0)
    assert np.allclose(foobaz(3.0), foobaz2(3.0))
    assert np.allclose(grad(foobaz)(3.0), grad(foobaz2)(3.0))


def checkpoint_memory():
    """This test is meant to be run manually, since it depends on
    memory_profiler and its behavior may vary."""
    try:
        from memory_profiler import memory_usage
    except ImportError:
        return

    def f(a):
        for _ in range(10):
            a = np.sin(a**2 + 1)
        return a

    checkpointed_f = checkpoint(f)

    def testfun(f, x):
        for _ in range(5):
            x = f(x)
        return np.sum(x)

    gradfun = grad(testfun, 1)

    A = npr.RandomState(0).randn(100000)
    max_usage = max(memory_usage((gradfun, (f, A))))
    max_checkpointed_usage = max(memory_usage((gradfun, (checkpointed_f, A))))

    assert max_checkpointed_usage < max_usage / 2.0


def test_make_jvp():
    A = npr.randn(3, 5)
    x = npr.randn(5)
    v = npr.randn(5)
    fun = lambda x: np.tanh(np.dot(A, x))

    jvp_explicit = lambda x: lambda v: np.dot(jacobian(fun)(x), v)
    jvp = make_jvp(fun)

    check_equivalent(jvp_explicit(x)(v), jvp(x)(v)[1])


def _make_explicit_ggnvp(f, g=lambda x: 1.0 / 2 * np.dot(x, x)):
    def ggnvp_maker(x):
        J = jacobian(f)(x)
        H = hessian(g)(f(x))

        def ggnvp(v):
            return np.dot(J.T, np.dot(H, np.dot(J, v)))

        return ggnvp

    return ggnvp_maker


def test_make_ggnvp():
    A = npr.randn(5, 4)
    x = npr.randn(4)
    v = npr.randn(4)

    fun = lambda x: np.dot(A, x)
    check_equivalent(make_ggnvp(fun)(x)(v), _make_explicit_ggnvp(fun)(x)(v))

    fun2 = lambda x: np.tanh(np.dot(A, x))
    check_equivalent(make_ggnvp(fun2)(x)(v), _make_explicit_ggnvp(fun2)(x)(v))


def test_make_ggnvp_nondefault_g():
    A = npr.randn(5, 4)
    x = npr.randn(4)
    v = npr.randn(4)

    g = lambda y: np.sum(2.0 * y**2 + y**4)

    fun = lambda x: np.dot(A, x)
    check_equivalent(make_ggnvp(fun, g)(x)(v), _make_explicit_ggnvp(fun, g)(x)(v))

    fun2 = lambda x: np.tanh(np.dot(A, x))
    check_equivalent(make_ggnvp(fun2, g)(x)(v), _make_explicit_ggnvp(fun2, g)(x)(v))


def test_grad_and_aux():
    A = npr.randn(5, 4)
    x = npr.randn(4)

    f = lambda x: (np.sum(np.dot(A, x)), x**2)
    g = lambda x: np.sum(np.dot(A, x))

    assert len(grad_and_aux(f)(x)) == 2

    check_equivalent(grad_and_aux(f)(x)[0], grad(g)(x))
    check_equivalent(grad_and_aux(f)(x)[1], x**2)


## No longer support this behavior
# def test_make_ggnvp_broadcasting():
#   A = npr.randn(4, 5)
#   x = npr.randn(10, 4)
#   v = npr.randn(10, 4)

#   fun = lambda x: np.tanh(np.dot(x, A))
#   res1 = np.stack([_make_explicit_ggnvp(fun)(xi)(vi) for xi, vi in zip(x, v)])
#   res2 = make_ggnvp(fun)(x)(v)
#   check_equivalent(res1, res2)


def test_wrapped_name_and_docs():
    def foo(x):
        pass

    assert grad.__name__ == "grad"
    # Python 3.13: Compiler now strip indents from docstrings.
    # https://docs.python.org/3.13/whatsnew/3.13.html#other-language-changes
    assert grad.__doc__.startswith(tuple(f"\n{indent}Returns a function which" for indent in ("    ", "")))
    assert grad(foo, 1).__name__ == "grad_of_foo_wrt_argnum_1"
    assert grad(foo, 1).__doc__.startswith("    grad of function foo with")



import warnings

from numpy_utils import combo_check

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import grad
from autograd.test_util import check_grads

npr.seed(1)


def test_numpy_version():
    import numpy

    assert np.__version__ == numpy.__version__


def test_dot():
    def fun(x, y):
        return np.dot(x, y)

    mat1 = npr.randn(10, 11)
    mat2 = npr.randn(10, 11)
    vect1 = npr.randn(10)
    vect2 = npr.randn(11)
    vect3 = npr.randn(11)

    check_grads(fun)(mat1, vect2)
    check_grads(fun)(mat1, mat2.T)
    check_grads(fun)(vect1, mat1)
    check_grads(fun)(vect2, vect3)


def test_dot_with_floats():
    def fun(x, y):
        return np.dot(x, y)

    mat1 = npr.randn(10, 11)
    vect1 = npr.randn(10)
    float1 = npr.randn()

    check_grads(fun)(mat1, float1)
    check_grads(fun)(float1, mat1)
    check_grads(fun)(vect1, float1)
    check_grads(fun)(float1, vect1)


# No longer supporting this
# def test_dot_method():
#     def fun(x, y): return x.dot(y)

#     mat1 = npr.randn(10, 11)
#     mat2 = npr.randn(10, 11)
#     vect1 = npr.randn(10)
#     vect2 = npr.randn(11)
#     vect3 = npr.randn(11)

#     check_grads(fun)(mat1, vect2)
#     check_grads(fun)(mat1, mat2.T)
#     check_grads(fun)(vect1, mat1)
#     check_grads(fun)(vect2, vect3)


def test_outer():
    def fun(x, y):
        return np.outer(x, y)

    vect2 = npr.randn(11)
    vect3 = npr.randn(11)

    check_grads(fun)(vect2, vect3)
    check_grads(fun)(vect2.T, vect3)
    check_grads(fun)(vect2.T, vect3.T)


def test_max():
    def fun(x):
        return np.max(x)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_max_axis():
    def fun(x):
        return np.max(x, axis=1)

    mat = npr.randn(3, 4, 5)
    check_grads(fun)(mat)


def test_max_axis_keepdims():
    def fun(x):
        return np.max(x, axis=1, keepdims=True)

    mat = npr.randn(3, 4, 5)
    check_grads(fun)(mat)


def test_min():
    def fun(x):
        return np.min(x)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_min_axis():
    def fun(x):
        return np.min(x, axis=1)

    mat = npr.randn(3, 4, 5)
    check_grads(fun)(mat)


def test_min_axis_keepdims():
    def fun(x):
        return np.min(x, axis=1, keepdims=True)

    mat = npr.randn(3, 4, 5)
    check_grads(fun)(mat)


def test_sum_1():
    def fun(x):
        return np.sum(x)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_sum_2():
    def fun(x):
        return np.sum(x, axis=0)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_sum_3():
    def fun(x):
        return np.sum(x, axis=0, keepdims=True)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_sum_with_axis_tuple():
    def fun(x):
        return np.sum(x, axis=(1, 2))

    mat = npr.randn(10, 11, 7)
    check_grads(fun)(mat)


def test_flipud():
    def fun(x):
        return np.flipud(x)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_fliplr():
    def fun(x):
        return np.fliplr(x)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_rot90():
    def fun(x):
        return np.rot90(x)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_cumsum_axis0():
    def fun(x):
        return np.cumsum(x, axis=0)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_cumsum_axis1():
    def fun(x):
        return np.cumsum(x, axis=1)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_cumsum_1d():
    def fun(x):
        return np.cumsum(x)

    mat = npr.randn(10)
    check_grads(fun)(mat)


def test_cumsum_no_axis():
    def fun(x):
        return np.cumsum(x)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_non_numpy_sum():
    def fun(x, y):
        return sum([x, y])

    mat1 = npr.randn(10, 11)
    mat2 = npr.randn(10, 11)
    check_grads(fun)(mat1, mat2)


def test_mean_1():
    def fun(x):
        return np.mean(x)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_mean_2():
    def fun(x):
        return np.mean(x, axis=0)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_mean_3():
    def fun(x):
        return np.mean(x, axis=0, keepdims=True)

    mat = npr.randn(10, 11)
    check_grads(fun)(mat)


def test_index_ints():
    A = npr.randn(5, 6, 4)

    def fun(x):
        return x[3, 0, 1]

    check_grads(fun)(A)


def test_index_slice():
    A = npr.randn(5, 6, 4)

    def fun(x):
        return x[::-1, 2:4, :]

    check_grads(fun)(A)


def test_index_lists():
    A = npr.randn(5, 6, 4)

    def fun(x):
        return x[[0, 1, 2], :, :]

    check_grads(fun)(A)


def test_index_mixed():
    A = npr.randn(5, 6, 4)

    def fun(x):
        return x[3, 2:, [1, 3]]

    check_grads(fun)(A)


def test_vector_slice():
    A = npr.randn(5)

    def fun(x):
        return x[2:4]

    check_grads(fun)(A)


def test_index_slice_fanout():
    A = npr.randn(5, 6, 4)

    def fun(x):
        y = x[::-1, 2:4, :]
        z = x[::-1, 3:5, :]
        return y + z

    check_grads(fun)(A)


def test_index_multiple_slices():
    A = npr.randn(7)

    def fun(x):
        y = x[2:6]
        z = y[1:3]
        return z

    check_grads(fun)(A)


def test_reshape_method():
    A = npr.randn(5, 6, 4)

    def fun(x):
        return x.reshape((5 * 4, 6))

    check_grads(fun)(A)


def test_reshape_call():
    A = npr.randn(5, 6, 4)

    def fun(x):
        return np.reshape(x, (5 * 4, 6))

    check_grads(fun)(A)


def test_reshape_method_nolist():
    # The reshape can be called in two different ways:
    # like A.reshape((5,4)) or A.reshape(5,4).
    # This test checks that we support the second way.
    A = npr.randn(5, 6, 4)

    def fun(x):
        return x.reshape(5 * 4, 6)

    check_grads(fun)(A)


def test_ravel_method():
    A = npr.randn(5, 6, 4)

    def fun(x):
        return x.ravel()

    check_grads(fun)(A)


def test_ravel_call():
    A = npr.randn(5, 6, 4)

    def fun(x):
        return np.ravel(x)

    check_grads(fun)(A)


def test_flatten_method():
    A = npr.randn(5, 6, 4)

    def fun(x):
        return x.flatten()

    check_grads(fun)(A)


def test_simple_append_list():
    A = [1.0, 2.0, 3.0]
    b = 4.0
    check_grads(np.append, argnum=(0, 1))(A, b)


def test_simple_append_arr():
    A = np.array([1.0, 2.0, 3.0])
    b = 4.0
    check_grads(np.append, argnum=(0, 1))(A, b)


def test_simple_append_list_2D():
    A = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
    B = [[7.0, 8.0, 9.0]]
    check_grads(np.append, argnum=(0, 1))(A, B, axis=0)


def test_simple_concatenate():
    A = npr.randn(5, 6, 4)
    B = npr.randn(4, 6, 4)

    def fun(x):
        return np.concatenate((A, x))

    check_grads(fun)(B)


def test_concatenate_axis_0():
    A = npr.randn(5, 6, 4)
    B = npr.randn(5, 6, 4)

    def fun(x):
        return np.concatenate((B, x, B))

    check_grads(fun)(A)


def test_concatenate_axis_1():
    A = npr.randn(5, 6, 4)
    B = npr.randn(5, 6, 4)

    def fun(x):
        return np.concatenate((B, x, B), axis=1)

    check_grads(fun)(A)


def test_concatenate_axis_1_unnamed():
    """Tests whether you can specify the axis without saying "axis=1"."""
    A = npr.randn(5, 6, 4)
    B = npr.randn(5, 6, 4)

    def fun(x):
        return np.concatenate((B, x, B), 1)

    check_grads(fun)(A)


def test_trace():
    def fun(x):
        return np.trace(x, offset=offset)

    mat = npr.randn(10, 11)
    offset = npr.randint(-9, 11)
    check_grads(fun)(mat)


def test_trace2():
    def fun(x):
        return np.trace(x, offset=offset)

    mat = npr.randn(11, 10)
    offset = npr.randint(-9, 11)
    check_grads(fun)(mat)


def test_trace_extradims():
    def fun(x):
        return np.trace(x, offset=offset)

    mat = npr.randn(5, 6, 4, 3)
    offset = npr.randint(-5, 6)
    check_grads(fun)(mat)


# TODO: Allow axis1, axis2 args.
# def test_trace_extradims2():
#     def fun(x): return np.trace(x, offset=offset, axis1=3,axis2=2)
#     mat = npr.randn(5,6,4,3)
#     offset = npr.randint(-5,6)
#     check_grads(fun)(mat)


def test_diag():
    def fun(x):
        return np.diag(x)

    mat = npr.randn(10, 10)
    check_grads(fun)(mat)


def test_transpose():
    def fun(x):
        return x.T

    mat = npr.randn(8, 8)
    check_grads(fun)(mat)


def test_roll():
    def fun(x):
        return np.roll(x, 2, axis=1)

    mat = npr.randn(4, 5)
    check_grads(fun)(mat)


def test_roll_no_axis():
    def fun(x):
        return np.roll(x, 2, axis=1)

    mat = npr.randn(4, 5)
    check_grads(fun)(mat)


def test_triu():
    def fun(x):
        return np.triu(x, k=2)

    mat = npr.randn(5, 5)
    check_grads(fun)(mat)


def test_tril():
    def fun(x):
        return np.tril(x, k=2)

    mat = npr.randn(5, 5)
    check_grads(fun)(mat)


def test_clip():
    def fun(x):
        return np.clip(x, a_min=0.1, a_max=1.1)

    mat = npr.randn(5, 5)
    check_grads(fun)(mat)


def test_prod_1():
    def fun(x):
        return np.prod(x)

    mat = npr.randn(2, 3) ** 2 / 10.0 + 0.1  # Gradient unstable when zeros are present.
    check_grads(fun)(mat)


def test_prod_2():
    def fun(x):
        return np.prod(x, axis=0)

    mat = npr.randn(2, 3) ** 2 + 0.1
    check_grads(fun)(mat)


def test_prod_3():
    def fun(x):
        return np.prod(x, axis=0, keepdims=True)

    mat = npr.randn(2, 3) ** 2 + 0.1
    check_grads(fun)(mat)


def test_prod_4():
    def fun(x):
        return np.prod(x)

    mat = npr.randn(7) ** 2 + 0.1
    check_grads(fun)(mat)


def test_1d_array():
    def fun(x):
        return np.array([x, x * 1.0, x + 2.5])

    check_grads(fun)(3.0)


def test_2d_array():
    def fun(x):
        return np.array([[x, x * 1.0, x + 2.5], [x**2, x, x / 2.0]])

    check_grads(fun)(3.0)


def test_1d_array_fanout():
    def fun(x):
        A = np.array([x, x * 1.0, x + 2.5])
        return A + A

    check_grads(fun)(3.0)


def test_2d_array_fanout():
    def fun(x):
        A = np.array([[x, x * 1.0, x + 2.5], [x**2, x, x / 2.0]])
        return A + A

    check_grads(fun)(3.0)


def test_array_from_scalar():
    def fun(x):
        return np.array(x)

    check_grads(fun)(3.0)


def test_array_from_arrays():
    def fun(x):
        return np.array([x, x])

    A = npr.randn(3, 2)
    check_grads(fun)(A)


def test_array_from_arrays_2():
    def fun(x):
        return np.array([[2 * x, x + 1], [x, x]])

    A = npr.randn(3, 2)
    check_grads(fun)(A)


def test_len():
    def fun(x):
        assert len(x) == 3
        return x

    A = npr.randn(3, 2)
    check_grads(fun)(A)


def test_r_basic():
    with warnings.catch_warnings(record=True) as w:

        def fun(x):
            c = npr.randn(3, 2)
            b = np.r_[x]
            return b

        A = npr.randn(3, 2)
        check_grads(fun)(A)


def test_r_double():
    with warnings.catch_warnings(record=True) as w:

        def fun(x):
            c = npr.randn(3, 2)
            b = np.r_[x, x]
            return b

        A = npr.randn(3, 2)
        check_grads(fun)(A)


def test_no_relation():
    with warnings.catch_warnings(record=True) as w:
        c = npr.randn(3, 2)

        def fun(x):
            return c

        A = npr.randn(3, 2)
        check_grads(fun)(A)


def test_r_no_relation():
    with warnings.catch_warnings(record=True) as w:
        c = npr.randn(3, 2)

        def fun(x):
            b = np.r_[c]
            return b

        A = npr.randn(3, 2)
        check_grads(fun)(A)


def test_r_node_and_const():
    with warnings.catch_warnings(record=True) as w:
        c = npr.randn(3, 2)

        def fun(x):
            b = np.r_[x, c]
            return b

        A = npr.randn(3, 2)
        check_grads(fun)(A)


def test_r_mixed():
    with warnings.catch_warnings(record=True) as w:
        c = npr.randn(3, 2)

        def fun(x):
            b = np.r_[x, c, x]
            return b

        A = npr.randn(3, 2)
        check_grads(fun)(A)


def test_r_slicing():
    with warnings.catch_warnings(record=True) as w:
        c = npr.randn(10)

        def fun(x):
            b = np.r_[x, c, 1:10]
            return b

        A = npr.randn(10)
        check_grads(fun)(A)


def test_c_():
    with warnings.catch_warnings(record=True) as w:
        c = npr.randn(3, 2)

        def fun(x):
            b = np.c_[x, c, x]
            return b

        A = npr.randn(3, 2)
        check_grads(fun)(A)


def test_c_mixed():
    with warnings.catch_warnings(record=True) as w:
        c = npr.randn(3, 2)

        def fun(x):
            b = np.c_[x, c, x]
            return b

        A = npr.randn(3, 2)
        check_grads(fun)(A)


def test_var_ddof():
    B = npr.randn(3)
    C = npr.randn(3, 4)
    D = npr.randn(1, 3)
    combo_check(np.var, (0,))([B, C, D], axis=[None], keepdims=[True, False], ddof=[0, 1])
    combo_check(np.var, (0,))([C, D], axis=[None, 1], keepdims=[True, False], ddof=[2])


def test_std_ddof():
    B = npr.randn(3)
    C = npr.randn(3, 4)
    D = npr.randn(1, 3)
    combo_check(np.std, (0,))([B, C, D], axis=[None], keepdims=[True, False], ddof=[0, 1])
    combo_check(np.std, (0,))([C, D], axis=[None, 1], keepdims=[True, False], ddof=[2])


def test_where():
    def fun(x, y):
        b = np.where(C, x, y)
        return b

    C = npr.randn(4, 5) > 0
    A = npr.randn(4, 5)
    B = npr.randn(4, 5)
    check_grads(fun)(A, B)


def test_squeeze_func():
    A = npr.randn(5, 1, 4)

    def fun(x):
        return np.squeeze(x)

    check_grads(fun)(A)


def test_squeeze_method():
    A = npr.randn(5, 1, 4)

    def fun(x):
        return x.squeeze()

    check_grads(fun)(A)


def test_repeat():
    A = npr.randn(5, 3, 4)

    def fun(x):
        return np.repeat(x, 2, axis=1)

    check_grads(fun)(A)


def test_repeat_axis1_rep1():
    A = npr.randn(5, 3, 4)

    def fun(x):
        return np.repeat(x, 1, axis=1)

    check_grads(fun)(A)


def test_repeat_axis0():
    A = npr.randn(5, 3)

    def fun(x):
        return np.repeat(x, 2, axis=0)

    check_grads(fun)(A)


def test_repeat_1d_axis0():
    A = npr.randn(5)

    def fun(x):
        return np.repeat(x, 2, axis=0)

    check_grads(fun)(A)


def test_repeat_axis0_rep1():
    A = npr.randn(5, 1)

    def fun(x):
        return np.repeat(x, 1, axis=0)

    check_grads(fun)(A)


def test_expand_dims():
    A = npr.randn(5, 1, 4)

    def fun(x):
        return np.expand_dims(x, 2)

    check_grads(fun)(A)


def test_tensordot_kwargs_by_position():
    def fun(x):
        return np.tensordot(x * np.ones((2, 2)), x * np.ones((2, 2)), 2)

    grad(fun)(1.0)


def test_multi_index():
    A = npr.randn(3)
    fun = lambda x: np.sum(x[[0, 0]])
    check_grads(fun)(A)


def test_multi_index2():
    A = npr.randn(3)
    fun = lambda x: np.sum(x[[0, 1, 0]])
    check_grads(fun)(A)


def test_index_dot_slices():
    A = npr.randn(4)

    def fun(x):
        return np.dot(x[:2], x[2:])

    check_grads(fun)(A)


# def test_index_exp_slicing():
#    def fun(x):
#        b = np.index_exp[x, x]
#        return b
#    A = npr.randn(10, 1)
#    check_grads(fun)(A)

# def test_s_slicing():
#    def fun(x):
#        b = np.s_[x, x]
#        return b
#    A = npr.randn(10, 1)
#    check_grads(fun)(A)

# TODO:
# getitem


def test_cast_to_int():
    inds = np.ones(5)[:, None]

    def fun(W):
        # glue W and inds together
        glued_together = np.concatenate((W, inds), axis=1)

        # separate W and inds back out
        new_W = W[:, :-1]
        new_inds = np.int64(W[:, -1])

        assert new_inds.dtype == np.int64
        return new_W[new_inds].sum()

    W = np.random.randn(5, 10)
    check_grads(fun)(W)


def test_make_diagonal():
    def fun(D):
        return np.make_diagonal(D, axis1=-1, axis2=-2)

    D = np.random.randn(4)
    A = np.make_diagonal(D, axis1=-1, axis2=-2)
    assert np.allclose(np.diag(A), D)
    check_grads(fun)(D)

    D = np.random.randn(3, 4)
    A = np.make_diagonal(D, axis1=-1, axis2=-2)
    assert all([np.allclose(np.diag(A[i]), D[i]) for i in range(3)])
    check_grads(fun)(D)


def test_diagonal():
    def fun(D):
        return np.diagonal(D, axis1=-1, axis2=-2)

    D = np.random.randn(4, 4)
    A = np.make_diagonal(D, axis1=-1, axis2=-2)
    check_grads(fun)(D)

    D = np.random.randn(3, 4, 4)
    A = np.make_diagonal(D, axis1=-1, axis2=-2)
    check_grads(fun)(D)


def test_nan_to_num():
    y = np.array([0.0, np.nan, np.inf, -np.inf])
    fun = lambda x: np.sum(np.sin(np.nan_to_num(x + y)))

    x = np.random.randn(4)
    check_grads(fun)(x)


# TODO(mattjj): np.frexp returns a pair of ndarrays and the second is an int
# type, for which there is currently no vspace registered
# def test_frexp():
#    fun = lambda x: np.frexp(x)[0]
#    A = 1.2 #np.random.rand(4,3) * 0.8 + 2.1
#    check_grads(fun)(A)


def test_max_equal_values():
    def fun(x):
        return np.max(np.array([x, x]))

    check_grads(fun)(1.0)


def test_max_equal_values_2d():
    def fun(x):
        return np.max(np.array([[x, x], [x, 0.5]]), axis=1)

    check_grads(fun)(1.0)
    check_grads(fun)(-1.0)


def test_min_3_way_equality():
    def fun(x):
        return np.min(np.array([[x, x, x], [x, 0.5, 0.5], [0.5, 0.5, 0.5], [x, x, 0.5]]), axis=0)

    check_grads(fun)(1.0)
    check_grads(fun)(-1.0)


def test_maximum_equal_values():
    def fun(x):
        return np.maximum(x, x)

    check_grads(fun)(1.0)


def test_maximum_equal_values_2d():
    def fun(x):
        return np.maximum(np.array([x, x, 0.5]), np.array([[x, 0.5, x], [x, x, 0.5]]))

    check_grads(fun)(1.0)
    check_grads(fun)(-1.0)
    check_grads(fun)(2.0)


def test_linspace():
    for num in [0, 1, 5]:

        def fun(x, y):
            return np.linspace(x, y, num)

        check_grads(fun)(1.2, 3.4)
        check_grads(fun)(1.2, -3.4)
        check_grads(fun)(1.2, 1.2)


def test_astype():
    x = np.arange(3, dtype="float32")

    def f(x):
        return np.sum(np.sin(x.astype("float64")))

    assert grad(f)(x).dtype == np.dtype("float32")


def test_gradient():
    check_grads(np.gradient, 0)(npr.randn(10))
    check_grads(np.gradient, 0)(npr.randn(10, 10))
    check_grads(np.gradient, 0)(npr.randn(10, 10, 10))

    for a in [None, 0, 1, -1, (0, 1), (0, -1)]:
        check_grads(np.gradient, 0)(npr.randn(10, 10, 10), axis=a)



# Implements auto-encoding variational Bayes.

from data import load_mnist, save_images

import autograd.numpy as np
import autograd.numpy.random as npr
import autograd.scipy.stats.norm as norm
from autograd import grad
from autograd.misc.optimizers import adam
from autograd.scipy.special import expit as sigmoid


def diag_gaussian_log_density(x, mu, log_std):
    return np.sum(norm.logpdf(x, mu, np.exp(log_std)), axis=-1)


def unpack_gaussian_params(params):
    # Params of a diagonal Gaussian.
    D = np.shape(params)[-1] // 2
    mean, log_std = params[:, :D], params[:, D:]
    return mean, log_std


def sample_diag_gaussian(mean, log_std, rs):
    return rs.randn(*mean.shape) * np.exp(log_std) + mean


def bernoulli_log_density(targets, unnormalized_logprobs):
    # unnormalized_logprobs are in R
    # Targets must be -1 or 1
    label_probabilities = -np.logaddexp(0, -unnormalized_logprobs * targets)
    return np.sum(label_probabilities, axis=-1)  # Sum across pixels.


def relu(x):
    return np.maximum(0, x)


def init_net_params(scale, layer_sizes, rs=npr.RandomState(0)):
    """Build a (weights, biases) tuples for all layers."""
    return [
        (
            scale * rs.randn(m, n),  # weight matrix
            scale * rs.randn(n),
        )  # bias vector
        for m, n in zip(layer_sizes[:-1], layer_sizes[1:])
    ]


def batch_normalize(activations):
    mbmean = np.mean(activations, axis=0, keepdims=True)
    return (activations - mbmean) / (np.std(activations, axis=0, keepdims=True) + 1)


def neural_net_predict(params, inputs):
    """Params is a list of (weights, bias) tuples.
    inputs is an (N x D) matrix.
    Applies batch normalization to every layer but the last."""
    for W, b in params[:-1]:
        outputs = batch_normalize(np.dot(inputs, W) + b)  # linear transformation
        inputs = relu(outputs)  # nonlinear transformation
    outW, outb = params[-1]
    outputs = np.dot(inputs, outW) + outb
    return outputs


def nn_predict_gaussian(params, inputs):
    # Returns means and diagonal variances
    return unpack_gaussian_params(neural_net_predict(params, inputs))


def generate_from_prior(gen_params, num_samples, noise_dim, rs):
    latents = rs.randn(num_samples, noise_dim)
    return sigmoid(neural_net_predict(gen_params, latents))


def p_images_given_latents(gen_params, images, latents):
    preds = neural_net_predict(gen_params, latents)
    return bernoulli_log_density(images, preds)


def vae_lower_bound(gen_params, rec_params, data, rs):
    # We use a simple Monte Carlo estimate of the KL
    # divergence from the prior.
    q_means, q_log_stds = nn_predict_gaussian(rec_params, data)
    latents = sample_diag_gaussian(q_means, q_log_stds, rs)
    q_latents = diag_gaussian_log_density(latents, q_means, q_log_stds)
    p_latents = diag_gaussian_log_density(latents, 0, 0)
    likelihood = p_images_given_latents(gen_params, data, latents)
    return np.mean(p_latents + likelihood - q_latents)


if __name__ == "__main__":
    # Model hyper-parameters
    latent_dim = 10
    data_dim = 784  # How many pixels in each image (28x28).
    gen_layer_sizes = [latent_dim, 300, 200, data_dim]
    rec_layer_sizes = [data_dim, 200, 300, latent_dim * 2]

    # Training parameters
    param_scale = 0.01
    batch_size = 200
    num_epochs = 15
    step_size = 0.001

    print("Loading training data...")
    N, train_images, _, test_images, _ = load_mnist()

    def binarise(images):
        on = images > 0.5
        images = images * 0 - 1
        images[on] = 1.0
        return images

    print("Binarising training data...")
    train_images = binarise(train_images)
    test_images = binarise(test_images)

    init_gen_params = init_net_params(param_scale, gen_layer_sizes)
    init_rec_params = init_net_params(param_scale, rec_layer_sizes)
    combined_init_params = (init_gen_params, init_rec_params)

    num_batches = int(np.ceil(len(train_images) / batch_size))

    def batch_indices(iter):
        idx = iter % num_batches
        return slice(idx * batch_size, (idx + 1) * batch_size)

    # Define training objective
    seed = npr.RandomState(0)

    def objective(combined_params, iter):
        data_idx = batch_indices(iter)
        gen_params, rec_params = combined_params
        return -vae_lower_bound(gen_params, rec_params, train_images[data_idx], seed) / data_dim

    # Get gradients of objective using autograd.
    objective_grad = grad(objective)

    print("     Epoch     |    Objective       |    Test ELBO  ")

    def print_perf(combined_params, iter, grad):
        if iter % 10 == 0:
            gen_params, rec_params = combined_params
            bound = np.mean(objective(combined_params, iter))
            message = f"{iter // num_batches:15}|{bound:20}|"
            if iter % 100 == 0:
                test_bound = -vae_lower_bound(gen_params, rec_params, test_images, seed) / data_dim
                message += f"{test_bound:20}"
            print(message)

            fake_data = generate_from_prior(gen_params, 20, latent_dim, seed)
            save_images(fake_data, "vae_samples.png", vmin=0, vmax=1)

    # The optimizers provided can optimize lists, tuples, or dicts of parameters.
    optimized_params = adam(
        objective_grad,
        combined_init_params,
        step_size=step_size,
        num_iters=num_epochs * num_batches,
        callback=print_perf,
    )



"""Gradients of the univariate t distribution."""

import scipy.stats

import autograd.numpy as np
from autograd.extend import defvjp, primitive
from autograd.numpy.numpy_vjps import unbroadcast_f
from autograd.scipy.special import psi

pdf = primitive(scipy.stats.t.pdf)
cdf = primitive(scipy.stats.t.cdf)
logpdf = primitive(scipy.stats.t.logpdf)
logcdf = primitive(scipy.stats.t.logcdf)


def grad_tlogpdf_diff(diff, df):
    return -diff * (1.0 + df) / (diff**2 + df)


def grad_tlogpdf_x(x, df, loc, scale):
    return grad_tlogpdf_diff((x - loc) / scale, df) / scale


def grad_tlogpdf_loc(x, df, loc, scale):
    return -grad_tlogpdf_diff((x - loc) / scale, df) / scale


def grad_tlogpdf_scale(x, df, loc, scale):
    diff = x - loc
    return -(df * (scale**2 - diff**2)) / (scale * (df * scale**2 + diff**2))


def grad_tlogpdf_df(x, df, loc, scale):
    y = (x - loc) / scale
    return 0.5 * (
        (y**2 * (df + 1)) / (df * (y**2 + df))
        - np.log(y**2 / df + 1)
        - 1.0 / df
        - psi(df / 2.0)
        + psi((df + 1) / 2.0)
    )


defvjp(
    pdf,
    lambda ans, x, df, loc=0.0, scale=1.0: unbroadcast_f(
        x, lambda g: g * ans * grad_tlogpdf_x(x, df, loc, scale)
    ),
    lambda ans, x, df, loc=0.0, scale=1.0: unbroadcast_f(
        df, lambda g: g * ans * grad_tlogpdf_df(x, df, loc, scale)
    ),
    lambda ans, x, df, loc=0.0, scale=1.0: unbroadcast_f(
        loc, lambda g: g * ans * grad_tlogpdf_loc(x, df, loc, scale)
    ),
    lambda ans, x, df, loc=0.0, scale=1.0: unbroadcast_f(
        scale, lambda g: g * ans * grad_tlogpdf_scale(x, df, loc, scale)
    ),
)

defvjp(
    cdf,
    lambda ans, x, df, loc=0.0, scale=1.0: unbroadcast_f(x, lambda g: g * pdf(x, df, loc, scale)),
    lambda ans, x, df, loc=0.0, scale=1.0: unbroadcast_f(loc, lambda g: -g * pdf(x, df, loc, scale)),
    argnums=(0, 2),
)

defvjp(
    logpdf,
    lambda ans, x, df, loc=0.0, scale=1.0: unbroadcast_f(x, lambda g: g * grad_tlogpdf_x(x, df, loc, scale)),
    lambda ans, x, df, loc=0.0, scale=1.0: unbroadcast_f(
        df, lambda g: g * grad_tlogpdf_df(x, df, loc, scale)
    ),
    lambda ans, x, df, loc=0.0, scale=1.0: unbroadcast_f(
        loc, lambda g: g * grad_tlogpdf_loc(x, df, loc, scale)
    ),
    lambda ans, x, df, loc=0.0, scale=1.0: unbroadcast_f(
        scale, lambda g: g * grad_tlogpdf_scale(x, df, loc, scale)
    ),
)

defvjp(
    logcdf,
    lambda ans, x, df, loc=0.0, scale=1.0: unbroadcast_f(
        x, lambda g: g * np.exp(logpdf(x, df, loc, scale) - logcdf(x, df, loc, scale))
    ),
    lambda ans, x, df, loc=0.0, scale=1.0: unbroadcast_f(
        loc, lambda g: -g * np.exp(logpdf(x, df, loc, scale) - logcdf(x, df, loc, scale))
    ),
    argnums=(0, 2),
)



import matplotlib.pyplot as plt

import autograd.numpy as np
import autograd.numpy.random as npr
import autograd.scipy.stats.multivariate_normal as mvn
import autograd.scipy.stats.norm as norm
from autograd import grad
from autograd.misc.optimizers import adam


def black_box_variational_inference(logprob, D, num_samples):
    """Implements http://arxiv.org/abs/1401.0118, and uses the
    local reparameterization trick from http://arxiv.org/abs/1506.02557"""

    def unpack_params(params):
        # Variational dist is a diagonal Gaussian.
        mean, log_std = params[:D], params[D:]
        return mean, log_std

    def gaussian_entropy(log_std):
        return 0.5 * D * (1.0 + np.log(2 * np.pi)) + np.sum(log_std)

    rs = npr.RandomState(0)

    def variational_objective(params, t):
        """Provides a stochastic estimate of the variational lower bound."""
        mean, log_std = unpack_params(params)
        samples = rs.randn(num_samples, D) * np.exp(log_std) + mean
        lower_bound = gaussian_entropy(log_std) + np.mean(logprob(samples, t))
        return -lower_bound

    gradient = grad(variational_objective)

    return variational_objective, gradient, unpack_params


if __name__ == "__main__":
    # Specify an inference problem by its unnormalized log-density.
    D = 2

    def log_density(x, t):
        mu, log_sigma = x[:, 0], x[:, 1]
        sigma_density = norm.logpdf(log_sigma, 0, 1.35)
        mu_density = norm.logpdf(mu, 0, np.exp(log_sigma))
        return sigma_density + mu_density

    # Build variational objective.
    objective, gradient, unpack_params = black_box_variational_inference(log_density, D, num_samples=2000)

    # Set up plotting code
    def plot_isocontours(ax, func, xlimits=[-2, 2], ylimits=[-4, 2], numticks=101):
        x = np.linspace(*xlimits, num=numticks)
        y = np.linspace(*ylimits, num=numticks)
        X, Y = np.meshgrid(x, y)
        zs = func(np.concatenate([np.atleast_2d(X.ravel()), np.atleast_2d(Y.ravel())]).T)
        Z = zs.reshape(X.shape)
        plt.contour(X, Y, Z)
        ax.set_yticks([])
        ax.set_xticks([])

    # Set up figure.
    fig = plt.figure(figsize=(8, 8), facecolor="white")
    ax = fig.add_subplot(111, frameon=False)
    plt.ion()
    plt.show(block=False)

    def callback(params, t, g):
        print(f"Iteration {t} lower bound {-objective(params, t)}")

        plt.cla()
        target_distribution = lambda x: np.exp(log_density(x, t))
        plot_isocontours(ax, target_distribution)

        mean, log_std = unpack_params(params)
        variational_contour = lambda x: mvn.pdf(x, mean, np.diag(np.exp(2 * log_std)))
        plot_isocontours(ax, variational_contour)
        plt.draw()
        plt.pause(1.0 / 30.0)

    print("Optimizing variational parameters...")
    init_mean = -1 * np.ones(D)
    init_log_std = -5 * np.ones(D)
    init_var_params = np.concatenate([init_mean, init_log_std])
    variational_params = adam(gradient, init_var_params, step_size=0.1, num_iters=2000, callback=callback)



from contextlib import contextmanager
from time import time

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import grad


@contextmanager
def tictoc(text=""):
    print("--- Start clock ---")
    t1 = time()
    yield
    dt = time() - t1
    print(f"--- Stop clock {text}: {dt} seconds elapsed ---")


def fan_out_fan_in():
    """The 'Pearlmutter test'"""

    def fun(x):
        for i in range(10**4):
            x = (x + x) / 2.0
        return np.sum(x)

    with tictoc():
        grad(fun)(1.0)


def convolution():
    # MNIST-scale convolution operation
    import autograd.scipy.signal

    convolve = autograd.scipy.signal.convolve
    dat = npr.randn(256, 3, 28, 28)
    kernel = npr.randn(3, 5, 5)
    with tictoc():
        convolve(dat, kernel, axes=([2, 3], [1, 2]), dot_axes=([1], [0]))


def dot_equivalent():
    # MNIST-scale convolution operation

    dat = npr.randn(256, 3, 24, 5, 24, 5)
    kernel = npr.randn(3, 5, 5)
    with tictoc():
        np.tensordot(dat, kernel, axes=[(1, 3, 5), (0, 1, 2)])


# fan_out_fan_in()
# convolution()
dot_equivalent()



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.
# http://asv.readthedocs.io/en/latest/writing_benchmarks.html
import autograd.numpy as np
from autograd import grad


class RNNSuite:
    """
    Checking speed on a vanilla RNN.
    """

    # NOTE: this is run each time we run a benchmark.
    # Might want to switch to setup_cache, which has to return an object which is loaded and unpacked in setup().
    def setup(self):
        self.batch_size = 16
        self.dtype = "float32"
        self.D = 2**10
        self.x = 0.01 * np.random.randn(self.batch_size, self.D).astype(self.dtype)
        self.W1 = 0.01 * np.random.randn(self.D, self.D).astype(self.dtype)
        self.b1 = 0.01 * np.random.randn(self.D).astype(self.dtype)
        self.Wout = 0.01 * np.random.randn(self.D, 1).astype(self.dtype)
        self.bout = 0.01 * np.random.randn(1).astype(self.dtype)
        self.l = (np.random.rand(self.batch_size, 1) > 0.5).astype(self.dtype)
        self.n = 50

        def autograd_rnn(params, x, label, n):
            W, b, Wout, bout = params
            h1 = x
            for i in range(n):
                h1 = np.tanh(np.dot(h1, W) + b)
            logit = np.dot(h1, Wout) + bout
            loss = -np.sum(label * logit - (logit + np.log(1 + np.exp(-logit))))
            return loss

        self.fn = autograd_rnn
        self.grad_fn = grad(self.fn)

    def rnn_grad(self):
        self.grad_fn((self.W1, self.b1, self.Wout, self.bout), self.x, self.l, self.n)

    def time_rnn_grad(self):
        self.rnn_grad()

    def peakmem_rnn_grad(self):
        self.rnn_grad()

    def time_manual_rnn_grad(self):
        self.manual_rnn_grad()

    def peakmem_manual_rnn_grad(self):
        self.manual_rnn_grad()

    def manual_rnn_grad(self):
        def repeat_to_match_shape(g, A, axis=None):
            gout = np.empty_like(A)
            if np.ndim(gout) == 0:
                gout = g
            else:
                gout = np.ones_like(A) * g
            return gout

        def sum_to_match_shape(sum_this, to_match_this):
            sum_this = np.sum(sum_this, axis=tuple(range(0, np.ndim(sum_this) - np.ndim(to_match_this))))
            for axis, size in enumerate(np.shape(to_match_this)):
                if size == 1:
                    sum_this = np.sum(sum_this, axis=axis, keepdims=True)
            return sum_this

        def grad_dot_A(g, A, B):
            ga = np.dot(g, B.T)
            ga = np.reshape(ga, np.shape(A))
            return ga

        def grad_dot_B(g, A, B):
            gb = np.dot(A.T, g)
            gb = np.reshape(gb, np.shape(B))
            return gb

        def _rnn_grad(x, W, b, Wout, bout, label, n):
            h1__1_stack, h1__1 = [], None
            h1__0_stack, h1__0 = [], None
            out_stack, out = [], None
            h1_stack = []
            h1 = x
            _for1 = list(range(n))

            for i in _for1:
                h1__1_stack.append(h1__1)
                h1__1 = np.dot(h1, W)
                h1__0_stack.append(h1__0)
                h1__0 = h1__1 + b
                h1_stack.append(h1)
                h1 = np.tanh(h1__0)
            out__0 = np.dot(h1, Wout)
            out = out__0 + bout
            loss__2 = label * out
            loss__7 = -out
            loss__6 = np.exp(loss__7)
            loss__5 = 1 + loss__6
            loss__4 = np.log(loss__5)
            loss__3 = out + loss__4
            loss__1 = loss__2 - loss__3

            # Begin Backward Pass
            g_loss = 1
            g_h1__0 = 0
            g_h1__1 = 0
            g_b = 0
            g_W = 0

            # Reverse of: loss = -loss__0
            g_loss__0 = -g_loss

            # Reverse of: loss__0 = np.sum(loss__1)
            g_loss__1 = repeat_to_match_shape(g_loss__0, loss__1)

            # Reverse of: loss__1 = loss__2 - loss__3
            g_loss__2 = sum_to_match_shape(g_loss__1, loss__2)
            g_loss__3 = sum_to_match_shape(-g_loss__1, loss__3)

            # Reverse of: loss__3 = out + loss__4
            g_out = sum_to_match_shape(g_loss__3, out)
            g_loss__4 = sum_to_match_shape(g_loss__3, loss__4)

            # Reverse of: loss__4 = np.log(loss__5)
            g_loss__5 = g_loss__4 / loss__5

            # Reverse of: loss__5 = 1 + loss__6
            g_loss__6 = sum_to_match_shape(g_loss__5, loss__6)

            # Reverse of: loss__6 = np.exp(loss__7)
            g_loss__7 = g_loss__6 * np.exp(loss__7)

            # Reverse of: loss__7 = -out
            g_out += -g_loss__7
            g_out += sum_to_match_shape(g_loss__2 * label, out)

            # Reverse of: out = out__0 + bout
            g_out__0 = sum_to_match_shape(g_out, out__0)
            g_bout = sum_to_match_shape(g_out, bout)

            # Reverse of: out__0 = np.dot(h1, Wout)
            g_h1 = grad_dot_A(g_out__0, h1, Wout)
            g_Wout = grad_dot_B(g_out__0, h1, Wout)
            _for1 = reversed(_for1)
            for i in _for1:
                h1 = h1_stack.pop()
                tmp_g0 = g_h1 / np.cosh(h1__0) ** 2.0
                g_h1 = 0
                g_h1__0 += tmp_g0
                h1__0 = h1__0_stack.pop()
                tmp_g1 = sum_to_match_shape(g_h1__0, h1__1)
                tmp_g2 = sum_to_match_shape(g_h1__0, b)
                g_h1__0 = 0
                g_h1__1 += tmp_g1
                g_b += tmp_g2
                h1__1 = h1__1_stack.pop()
                tmp_g3 = grad_dot_A(g_h1__1, h1, W)
                tmp_g4 = grad_dot_B(g_h1__1, h1, W)
                g_h1__1 = 0
                g_h1 += tmp_g3
                g_W += tmp_g4
            return g_W, g_b, g_Wout, g_bout

        _rnn_grad(self.x, self.W1, self.b1, self.Wout, self.bout, self.l, self.n)
        pass



import scipy.optimize

import autograd.numpy as np
import autograd.numpy.random as npr
from autograd import grad
from autograd.scipy.special import gammaln

# The code in this example implements a method for finding a stationary point of
# the negative binomial likelihood via Newton's method, described here:
# https://en.wikipedia.org/wiki/Negative_binomial_distribution#Maximum_likelihood_estimation


def newton(f, x0):
    # wrap scipy.optimize.newton with our automatic derivatives
    return scipy.optimize.newton(f, x0, fprime=grad(f), fprime2=grad(grad(f)))


def negbin_loglike(r, p, x):
    # the negative binomial log likelihood we want to maximize
    return gammaln(r + x) - gammaln(r) - gammaln(x + 1) + x * np.log(p) + r * np.log(1 - p)


def negbin_sample(r, p, size):
    # a negative binomial is a gamma-compound-Poisson
    return npr.poisson(npr.gamma(r, p / (1 - p), size=size))


def fit_maxlike(x, r_guess):
    # follows Wikipedia's section on negative binomial max likelihood
    assert np.var(x) > np.mean(x), "Likelihood-maximizing parameters don't exist!"
    loglike = lambda r, p: np.sum(negbin_loglike(r, p, x))
    p = lambda r: np.sum(x) / np.sum(r + x)
    rprime = lambda r: grad(loglike)(r, p(r))
    r = newton(rprime, r_guess)
    return r, p(r)


if __name__ == "__main__":
    # generate data
    npr.seed(0)
    data = negbin_sample(r=5, p=0.5, size=1000)

    # fit likelihood-extremizing parameters
    r, p = fit_maxlike(data, r_guess=1)

    # report fit
    print("Fit parameters:")
    print(f"r={r}, p={p}")

    print("Check that we are at a local stationary point:")
    loglike = lambda r, p: np.sum(negbin_loglike(r, p, data))
    grad_both = grad(loglike, argnum=(0, 1))
    print(grad_both(r, p))

    import matplotlib.pyplot as plt

    xm = data.max()
    plt.figure()
    plt.hist(data, bins=np.arange(xm + 1) - 0.5, normed=True, label="normed data counts")
    plt.xlim(0, xm)
    plt.plot(np.arange(xm), np.exp(negbin_loglike(r, p, np.arange(xm))), label="maxlike fit")
    plt.xlabel("k")
    plt.ylabel("p(k)")
    plt.legend(loc="best")
    plt.show()



# This file is to check that future division works.

from test_binary_ops import arg_pairs

import autograd.numpy as np
from autograd.test_util import check_grads


def test_div():
    fun = lambda x, y: x / y
    make_gap_from_zero = lambda x: np.sqrt(x**2 + 0.5)
    for arg1, arg2 in arg_pairs():
        arg1 = make_gap_from_zero(arg1)
        arg2 = make_gap_from_zero(arg2)
        check_grads(fun)(arg1, arg2)
